/*
 * Copyright (c) 2011 Douglas Gilbert.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

/************************************************************************
 * g20gpio_set.c
 * 
 * Utility for setting AT91SAM9G20 GPIO line attributes.
 * The G20 has 3 banks of 32 gpio lines: PA0-PA31, PB0-PB31 and PC0-PC31.
 * Uses AT91SAM9G20 memory mapped IO.
 *
 ****************************************************/

#define _XOPEN_SOURCE 500

#include <stdio.h>     
#include <stdlib.h>     
#include <ctype.h>     
#include <string.h>     
#include <unistd.h>    
#include <sys/mman.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>


static const char * version_str = "1.00 20110829";


#define GPIO_BANKS_G20 3    /* PA0-31,PB0-31 and PC0-31 */
#define LINES_PER_BANK 32
#define MAP_SIZE 4096   /* assume to be power of 2 */
#define MAP_MASK (MAP_SIZE - 1)
#define DEV_MEM "/dev/mem"


struct mmap_state {
    void * mmap_ptr;
    off_t prev_mask_addrp;
    int mmap_ok;
};

static unsigned int pio_per[] = {0xfffff400, 0xfffff600, 0xfffff800};
static unsigned int pio_pdr[] = {0xfffff404, 0xfffff604, 0xfffff804};
static unsigned int pio_oer[] = {0xfffff410, 0xfffff610, 0xfffff810};
static unsigned int pio_odr[] = {0xfffff414, 0xfffff614, 0xfffff814};
static unsigned int pio_ifer[] = {0xfffff420, 0xfffff620, 0xfffff820};
static unsigned int pio_ifdr[] = {0xfffff424, 0xfffff624, 0xfffff824};
static unsigned int pio_ier[] = {0xfffff440, 0xfffff640, 0xfffff840};
static unsigned int pio_idr[] = {0xfffff444, 0xfffff644, 0xfffff844};
static unsigned int pio_mder[] = {0xfffff450, 0xfffff650, 0xfffff850};
static unsigned int pio_mddr[] = {0xfffff454, 0xfffff654, 0xfffff854};
static unsigned int pio_puer[] = {0xfffff464, 0xfffff664, 0xfffff864};
static unsigned int pio_pudr[] = {0xfffff460, 0xfffff660, 0xfffff860};
static unsigned int pio_sodr[] = {0xfffff430, 0xfffff630, 0xfffff830};
static unsigned int pio_codr[] = {0xfffff434, 0xfffff634, 0xfffff834};
static unsigned int pio_asr[] = {0xfffff470, 0xfffff670, 0xfffff870};
static unsigned int pio_bsr[] = {0xfffff474, 0xfffff674, 0xfffff874};

static int verbose = 0;


static void
usage(void)
{
    fprintf(stderr, "Usage: "
            "g20gpio_status [-A] [-b <bn>] [-B] [-e] [-g] [-G] [-h] [-i] "
            "[-I]\n"
            "                      [-m] [-M] [-o] [-O] [-P] [-p <port>] "
            "[-S <n>] [-u]\n"
            "                      [-U] [-v] [-V]\n"
            "  where:\n"
            "    -A           set line to 'peripheral A' (excludes '-B' "
            "and '-P')\n"
            "    -b <bn>      bit number within port (0 to 31)\n"
            "    -B           set line to 'peripheral B' (excludes '-A' "
            "and '-P')\n"
            "    -e           enumerate pin names with corresponding "
            "kernel pin\n"
            "    -g           disables glitch input filter (excludes '-G')\n"
            "    -G           enables glitch input filter (excludes '-g')\n"
            "    -i           interrupt disable (excludes '-I')\n"
            "    -I           interrupt enable (excludes '-i')\n"
            "    -h           print usage message\n"
            "    -m           disables multi-drive (open drain) (excludes "
            "'-M')\n"
            "    -M           enables multi-drive (open drain) (excludes "
            "'-m')\n"
            "    -o           output disable (PIO mode only) (excludes "
            "'-O')\n"
            "    -O           output enable (PIO mode only) (excludes "
            "'-o')\n"
            "    -p <port>    port bank ('A' to 'C') or gpio kernel line "
            "number\n"
            "    -P           set line to 'PIO' mode (excludes '-A' "
            "and '-B')\n"
            "    -S <n>       set output data line to <n> (either 0 or 1)\n"
            "    -u           disables internal pull-up (excludes '-U')\n"
            "    -U           enables internal pull-up (excludes '-u')\n"
            "    -v           increase verbosity (multiple times for more)\n"
            "    -V           print version string then exit\n\n"
            "Set AT91SAM9G20 GPIO attributes. Uses memory mapped "
            "IO to access\nPIO registers directly. Bypasses kernel.\n");
}

/* Since we map a whole page (MAP_SIZE) then that page may already be
 * mapped which means we can skip a munmap() and mmap() call. Returns
 * new or re-used pointer to mmapped page, or returns NULL if problem. */
static void *
check_mmap(int mem_fd, unsigned int wanted_addr, struct mmap_state * msp)
{
    off_t mask_addr;

    mask_addr = (wanted_addr & ~MAP_MASK);
    if ((0 == msp->mmap_ok) || (msp->prev_mask_addrp != mask_addr)) {
        if (msp->mmap_ok) {
            if (-1 == munmap(msp->mmap_ptr, MAP_SIZE)) {
                fprintf(stderr, "mmap_ptr=%p:\n", msp->mmap_ptr);
                perror("    munmap");
                return NULL;
            } else if (verbose > 2)
                fprintf(stderr, "munmap() ok, mmap_ptr=%p\n", msp->mmap_ptr);
        }
        msp->mmap_ptr = mmap(0, MAP_SIZE, PROT_READ | PROT_WRITE, MAP_SHARED,
                             mem_fd, mask_addr);
        if ((void *)-1 == msp->mmap_ptr) {
            msp->mmap_ok = 0;
            fprintf(stderr, "addr=0x%x, mask_addr=0x%lx :\n", wanted_addr,
                    mask_addr);
            perror("    mmap");
            return NULL;
        }
        msp->mmap_ok = 1;
        msp->prev_mask_addrp = mask_addr;
        if (verbose > 2)
            fprintf(stderr, "mmap() ok, mask_addr=0x%lx, mmap_ptr=%p\n",
                    mask_addr, msp->mmap_ptr);
    }
    return msp->mmap_ptr;
}



int
main(int argc, char ** argv)
{
    int opt, k, j, n, num, mem_fd, pioc_num;
    int peri_a = 0;
    int peri_b = 0;
    int do_pio = 0;
    int enumerate = 0;
    int en_output = 0;
    int di_output = 0;
    int en_glitch = 0;
    int di_glitch = 0;
    int en_interrupt = 0;
    int di_interrupt = 0;
    int en_multi = 0;
    int di_multi = 0;
    int en_pullup = 0;
    int di_pullup = 0;
    int out_data = -1;
    int knum = -1;
    int bit_num = -1;
    unsigned int bit_mask;
    unsigned int addr;
    char ch;
    char bank = '\0';
    void * ap;
    void * mmap_ptr = (void *)-1;
    struct mmap_state mstat;

    while ((opt = getopt(argc, argv, "Ab:BegGhiImMoOp:PS:uUvV")) != -1) {
        switch (opt) {
        case 'A':
            ++peri_a;
            break;
        case 'b':
            k = atoi(optarg);
            if ((k < 0) || (k > 31)) {
                fprintf(stderr, "'-b' expects a bit number from 0 to 31\n");
                return 1;
            }
            bit_num = k;
            break;
        case 'B':
            ++peri_b;
            break;
        case 'e':
            ++enumerate;
            break;
        case 'g':
            ++di_glitch;
            break;
        case 'G':
            ++en_glitch;
            break;
        case 'h':
            usage();
            return 0;
            break;
        case 'i':
            ++di_interrupt;
            break;
        case 'I':
            ++en_interrupt;
            break;
        case 'm':
            ++di_multi;
            break;
        case 'M':
            ++en_multi;
            break;
        case 'o':
            ++di_output;
            break;
        case 'O':
            ++en_output;
            break;
        case 'p':
            if (isalpha(*optarg)) {
                ch = toupper(*optarg);
                if ((ch >= 'A') && (ch <= 'C'))
                    bank = ch;
                else {
                    fprintf(stderr, "'-p' expects a letter ('A' to 'C')\n");
                    return 1;
                }
            } else if (isdigit(*optarg)) {
                k = atoi(optarg);
                if ((k < 32) || (k > 127)) {
                    fprintf(stderr, "'-p' expects a letter or a number "
                            "32 or greater\n");
                    return 1;
                }
                knum = k;
            } else {
                fprintf(stderr, "'-p' expects a letter ('A' to 'C') or "
                        "a number\n");
                return 1;
            }
            break;
        case 'P':
            ++do_pio;
            break;
        case 'S':
            k = atoi(optarg);
            if ((k < 0) || (k > 1)) {
                fprintf(stderr, "'-S' expects a value of 0 or 1\n");
                return 1;
            }
            out_data = k;
            break;
        case 'u':
            ++di_pullup;
            break;
        case 'U':
            ++en_pullup;
            break;
        case 'v':
            ++verbose;
            break;
        case 'V':
            printf("%s\n", version_str);
            return 0;
        default: /* '?' */
            usage();
            return 1;
        }
    }
    if (optind < argc) {
        if (optind < argc) {
            for (; optind < argc; ++optind)
                fprintf(stderr, "Unexpected extra argument: %s\n",
                        argv[optind]);
            usage();
            return 1;
        }
    }

    if (enumerate) {
        num = GPIO_BANKS_G20;
        for (k = 0; k < LINES_PER_BANK; ++k) {
            for (j = 0; j < num; ++j) {
                n = ((j + 1) * 32) + k;
                printf("%sP%c%d: %d   ", (j ? "\t" : ""), 'A' + j, k, n);
            }
            printf("\n");
        }
        return 0;
    }

    if (knum >= 0) {
        if (bit_num >= 0) {
            fprintf(stderr, "Give either '-p <knum>' or ('-b <bn>' and "
                    "'-p <bank>') but not both\n");
            return 1;
        }
    } else if (bank) {
        if (bit_num < 0) {
            fprintf(stderr, "If '-p <bank>' given then also need "
                    "'-b <bn>'\n");
            return 1;
        } else
            knum = ((bank + 1 - 'A') * 32) + bit_num;
    } else {
        fprintf(stderr, "Need to give gpio line with '-p <port>' and/or "
                "'-b <bn>'\n");
        usage();
        return 1;
    }
    if ((!!peri_a + !!peri_b + !!do_pio) > 1) {
        fprintf(stderr, "Can only have one of '-A', '-B' and '-P'\n");
        usage();
        return 1;
    }
    if ((!!di_glitch + !!en_glitch) > 1) {
        fprintf(stderr, "Can only have one of '-g' and '-G'\n");
        usage();
        return 1;
    }
    if ((!!di_interrupt + !!en_interrupt) > 1) {
        fprintf(stderr, "Can only have one of '-i' and '-I'\n");
        usage();
        return 1;
    }
    if ((!!di_multi + !!en_multi) > 1) {
        fprintf(stderr, "Can only have one of '-m' and '-M'\n");
        usage();
        return 1;
    }
    if ((!!di_output + !!en_output) > 1) {
        fprintf(stderr, "Can only have one of '-o' and '-O'\n");
        usage();
        return 1;
    }
    if ((!!di_pullup + !!en_pullup) > 1) {
        fprintf(stderr, "Can only have one of '-u' and '-U'\n");
        usage();
        return 1;
    }

    pioc_num = bank ? (bank - 'A') : ((knum - 32) / 32);
    if (bit_num < 0)
        bit_num = knum % 32;
    bit_mask = 1 << bit_num;
    if (verbose)
        printf("P%c%d:\n", 'A' + pioc_num, bit_num);
    if (verbose > 1)
        printf("  bit_mask=0x%08x\n", bit_mask);

    if ((mem_fd = open(DEV_MEM, O_RDWR | O_SYNC)) < 0) {
        perror("open of " DEV_MEM " failed");
        return 1;
    } else if (verbose > 2)
        printf("open(" DEV_MEM "O_RDWR | O_SYNC) okay\n");

    memset(&mstat, 0, sizeof(mstat));

    if (peri_a || peri_b || do_pio) {
        if (do_pio) {
            mmap_ptr = check_mmap(mem_fd, pio_per[pioc_num], &mstat);
            if (NULL == mmap_ptr)
                return 1;
            ap = mmap_ptr + (pio_per[pioc_num] & MAP_MASK);
            *((unsigned int *)ap) = bit_mask;
        } else {
            addr = (peri_a ? pio_asr[pioc_num] : pio_bsr[pioc_num]);
            mmap_ptr = check_mmap(mem_fd, addr, &mstat);
            if (NULL == mmap_ptr)
                return 1;
            ap = mmap_ptr + (addr & MAP_MASK);
            *((unsigned int *)ap) = bit_mask;

            mmap_ptr = check_mmap(mem_fd, pio_pdr[pioc_num], &mstat);
            if (NULL == mmap_ptr)
                return 1;
            ap = mmap_ptr + (pio_pdr[pioc_num] & MAP_MASK);
            *((unsigned int *)ap) = bit_mask;
        }
    }
    if (di_glitch || en_glitch) {
        addr = (di_glitch ? pio_ifdr[pioc_num] : pio_ifer[pioc_num]);
        mmap_ptr = check_mmap(mem_fd, addr, &mstat);
        if (NULL == mmap_ptr)
            return 1;
        ap = mmap_ptr + (addr & MAP_MASK);
        *((unsigned int *)ap) = bit_mask;
    }
    if (di_interrupt || en_interrupt) {
        addr = (di_interrupt ? pio_idr[pioc_num] : pio_ier[pioc_num]);
        mmap_ptr = check_mmap(mem_fd, addr, &mstat);
        if (NULL == mmap_ptr)
            return 1;
        ap = mmap_ptr + (addr & MAP_MASK);
        *((unsigned int *)ap) = bit_mask;
    }
    if (di_multi || en_multi) {
        addr = (di_multi ? pio_mddr[pioc_num] : pio_mder[pioc_num]);
        mmap_ptr = check_mmap(mem_fd, addr, &mstat);
        if (NULL == mmap_ptr)
            return 1;
        ap = mmap_ptr + (addr & MAP_MASK);
        *((unsigned int *)ap) = bit_mask;
    }
    if (di_output || en_output) {
        addr = (di_output ? pio_odr[pioc_num] : pio_oer[pioc_num]);
        mmap_ptr = check_mmap(mem_fd, addr, &mstat);
        if (NULL == mmap_ptr)
            return 1;
        ap = mmap_ptr + (addr & MAP_MASK);
        *((unsigned int *)ap) = bit_mask;
    }
    if (di_pullup || en_pullup) {
        addr = (di_pullup ? pio_pudr[pioc_num] : pio_puer[pioc_num]);
        mmap_ptr = check_mmap(mem_fd, addr, &mstat);
        if (NULL == mmap_ptr)
            return 1;
        ap = mmap_ptr + (addr & MAP_MASK);
        *((unsigned int *)ap) = bit_mask;
    }
    if (out_data >= 0) {
        addr = (out_data ? pio_sodr[pioc_num] : pio_codr[pioc_num]);
        mmap_ptr = check_mmap(mem_fd, addr, &mstat);
        if (NULL == mmap_ptr)
            return 1;
        ap = mmap_ptr + (addr & MAP_MASK);
        *((unsigned int *)ap) = bit_mask;
    }

    if (mstat.mmap_ok) {
        if (-1 == munmap(mmap_ptr, MAP_SIZE)) {
            fprintf(stderr, "mmap_ptr=%p:\n", mmap_ptr);
            perror("    munmap");
            return 1;
        } else if (verbose > 2)
            fprintf(stderr, "trailing munmap() ok, mmap_ptr=%p\n", mmap_ptr);
    }
    if (mem_fd >= 0)
        close(mem_fd);
    return 0;
}
