/*
 * Copyright (c) 2011 Douglas Gilbert.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

/************************************************************************
 * g20gpio_status.c
 * 
 * Utility for fetching AT91SAM9G20 GPIO status values.
 * The G20 has 3 banks of 32 gpio lines: PA0-PA31, PB0-PB31 and PC0-PC31.
 * Uses AT91SAM9G20 memory mapped IO.
 *
 ****************************************************/

#define _XOPEN_SOURCE 500

#include <stdio.h>     
#include <stdlib.h>     
#include <ctype.h>     
#include <string.h>     
#include <unistd.h>    
#include <sys/mman.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>


static const char * version_str = "1.00 20110829";


#define GPIO_BANKS_G20 3    /* PA0-31,PB0-31 and PC0-31 */
#define LINES_PER_BANK 32
#define MAP_SIZE 4096   /* assume to be power of 2 */
#define MAP_MASK (MAP_SIZE - 1)
#define DEV_MEM "/dev/mem"


struct mmap_state {
    void * mmap_ptr;
    off_t prev_mask_addrp;
    int mmap_ok;
};

static unsigned int pio_psr[] = {0xfffff408, 0xfffff608, 0xfffff808};
static unsigned int pio_osr[] = {0xfffff418, 0xfffff618, 0xfffff818};
static unsigned int pio_ifsr[] = {0xfffff428, 0xfffff628, 0xfffff828};
static unsigned int pio_odsr[] = {0xfffff438, 0xfffff638, 0xfffff838};
static unsigned int pio_pdsr[] = {0xfffff43c, 0xfffff63c, 0xfffff83c};
static unsigned int pio_imr[] = {0xfffff448, 0xfffff648, 0xfffff848};
static unsigned int pio_isr[] = {0xfffff44c, 0xfffff64c, 0xfffff84c};
static unsigned int pio_mdsr[] = {0xfffff458, 0xfffff658, 0xfffff858};
static unsigned int pio_pusr[] = {0xfffff468, 0xfffff668, 0xfffff868};
static unsigned int pio_absr[] = {0xfffff478, 0xfffff678, 0xfffff878};
static unsigned int pio_owsr[] = {0xfffff4a8, 0xfffff6a8, 0xfffff8a8};

static int verbose = 0;


static void
usage(void)
{
    fprintf(stderr, "Usage: "
            "g20gpio_status [-b <bn>] [-B] [-e] [-h] [-i] [-p <port>] "
            "[-v] [-V]\n"
            "  where:\n"
            "    -b <bn>      bit number within port (0 to 31)\n"
            "    -B           one line summary from registers for "
            "given line\n"
            "                 use twice for <bn> prefix without imr, "
            "isr and owsr\n"
            "    -e           enumerate pin names with corresponding "
            "kernel pin\n"
            "    -h           print usage message\n"
            "    -i           read interrupt status register which clears "
            "it\n"
            "    -p <port>    port bank ('A' to 'C') or gpio kernel line "
            "number\n"
            "                 32 for PA0, 33 for PA1 ... 127 for PC31\n"
            "    -v           increase verbosity (multiple times for more)\n"
            "    -V           print version string then exit\n\n"
            "AT91SAM9G20 GPIO fetch status program. Uses memory mapped "
            "IO to fetch\nPIO registers and shows setting for given line.\n");
}

/* Since we map a whole page (MAP_SIZE) then that page may already be
 * mapped which means we can skip a munmap() and mmap() call. Returns
 * new or re-used pointer to mmapped page, or returns NULL if problem. */
static void *
check_mmap(int mem_fd, unsigned int wanted_addr, struct mmap_state * msp)
{
    off_t mask_addr;

    mask_addr = (wanted_addr & ~MAP_MASK);
    if ((0 == msp->mmap_ok) || (msp->prev_mask_addrp != mask_addr)) {
        if (msp->mmap_ok) {
            if (-1 == munmap(msp->mmap_ptr, MAP_SIZE)) {
                fprintf(stderr, "mmap_ptr=%p:\n", msp->mmap_ptr);
                perror("    munmap");
                return NULL;
            } else if (verbose > 2)
                fprintf(stderr, "munmap() ok, mmap_ptr=%p\n", msp->mmap_ptr);
        }
        msp->mmap_ptr = mmap(0, MAP_SIZE, PROT_READ | PROT_WRITE, MAP_SHARED,
                             mem_fd, mask_addr);
        if ((void *)-1 == msp->mmap_ptr) {
            msp->mmap_ok = 0;
            fprintf(stderr, "addr=0x%x, mask_addr=0x%lx :\n", wanted_addr,
                    mask_addr);
            perror("    mmap");
            return NULL;
        }
        msp->mmap_ok = 1;
        msp->prev_mask_addrp = mask_addr;
        if (verbose > 2)
            fprintf(stderr, "mmap() ok, mask_addr=0x%lx, mmap_ptr=%p\n",
                    mask_addr, msp->mmap_ptr);
    }
    return msp->mmap_ptr;
}



int
main(int argc, char ** argv)
{
    int opt, k, j, n, num, mem_fd, pioc_num, psr, osr, ifsr, odsr, pdsr;
    int imr, isr, mdsr, pusr, absr, owsr;
    int brief = 0;
    int enumerate = 0;
    int interrupt = 0;
    int knum = -1;
    int bit_num = -1;
    unsigned int bit_mask;
    char ch;
    char bank = '\0';
    void * ap;
    void * mmap_ptr = (void *)-1;
    struct mmap_state mstat;

    while ((opt = getopt(argc, argv, "b:Behip:vV")) != -1) {
        switch (opt) {
        case 'b':
            k = atoi(optarg);
            if ((k < 0) || (k > 31)) {
                fprintf(stderr, "'-b' expects a bit number from 0 to 31\n");
                exit(EXIT_FAILURE);
            }
            bit_num = k;
            break;
        case 'B':
            ++brief;
            break;
        case 'e':
            ++enumerate;
            break;
        case 'h':
            usage();
            exit(EXIT_SUCCESS);
            break;
        case 'i':
            ++interrupt;
            break;
        case 'p':
            if (isalpha(*optarg)) {
                ch = toupper(*optarg);
                if ((ch >= 'A') && (ch <= 'C'))
                    bank = ch;
                else {
                    fprintf(stderr, "'-p' expects a letter ('A' to 'C')\n");
                    exit(EXIT_FAILURE);
                }
            } else if (isdigit(*optarg)) {
                k = atoi(optarg);
                if ((k < 32) || (k > 127)) {
                    fprintf(stderr, "'-p' expects a letter or a number "
                            "32 or greater\n");
                    exit(EXIT_FAILURE);
                }
                knum = k;
            } else {
                fprintf(stderr, "'-p' expects a letter ('A' to 'C') or "
                        "a number\n");
                exit(EXIT_FAILURE);
            }
            break;
        case 'v':
            ++verbose;
            break;
        case 'V':
            printf("%s\n", version_str);
            exit(EXIT_SUCCESS);
        default: /* '?' */
            usage();
            exit(EXIT_FAILURE);
        }
    }
    if (optind < argc) {
        if (optind < argc) {
            for (; optind < argc; ++optind)
                fprintf(stderr, "Unexpected extra argument: %s\n",
                        argv[optind]);
            usage();
            exit(EXIT_FAILURE);
        }
    }

    if (enumerate) {
        num = GPIO_BANKS_G20;
        for (k = 0; k < LINES_PER_BANK; ++k) {
            for (j = 0; j < num; ++j) {
                n = ((j + 1) * 32) + k;
                printf("%sP%c%d: %d   ", (j ? "\t" : ""), 'A' + j, k, n);
            }
            printf("\n");
        }
        return 0;
    }

    if (knum >= 0) {
        if (bit_num >= 0) {
            fprintf(stderr, "Give either '-p <knum>' or ('-b <bn>' and "
                    "'-p <bank>') but not both\n");
            exit(EXIT_FAILURE);
        }
    } else if (bank) {
        if (bit_num < 0) {
            fprintf(stderr, "If '-p <bank>' given then also need "
                    "'-b <bn>'\n");
            exit(EXIT_FAILURE);
        } else
            knum = ((bank + 1 - 'A') * 32) + bit_num;
    } else {
        fprintf(stderr, "Need to give gpio line with '-p <port>' and/or "
                "'-b <bn>'\n");
        usage();
        exit(EXIT_FAILURE);
    }
    pioc_num = bank ? (bank - 'A') : ((knum - 32) / 32);
    if (bit_num < 0)
        bit_num = knum % 32;
    bit_mask = 1 << bit_num;
    if (verbose)
        printf("P%c%d:\n", 'A' + pioc_num, bit_num);
    if (verbose > 1)
        printf("  bit_mask=0x%08x\n", bit_mask);

    if ((mem_fd = open(DEV_MEM, O_RDWR | O_SYNC)) < 0) {
        perror("open of " DEV_MEM " failed");
        return 1;
    } else if (verbose > 2)
        printf("open(" DEV_MEM "O_RDWR | O_SYNC) okay\n");

    memset(&mstat, 0, sizeof(mstat));

    mmap_ptr = check_mmap(mem_fd, pio_psr[pioc_num], &mstat);
    if (NULL == mmap_ptr)
        return 1;
    ap = mmap_ptr + (pio_psr[pioc_num] & MAP_MASK);
    psr = !!( *((unsigned int *)ap) & bit_mask);
    if (0 == brief)
        printf("  PIO status: %d (%s)\n", psr,
               (psr ? "PIO ACTIVE, peripheral inactive" :
                      "peripheral ACTIVE"));

    mmap_ptr = check_mmap(mem_fd, pio_osr[pioc_num], &mstat);
    if (NULL == mmap_ptr)
        return 1;
    ap = mmap_ptr + (pio_osr[pioc_num] & MAP_MASK);
    osr = !!( *((unsigned int *)ap) & bit_mask);
    if (0 == brief)
        printf("  PIO output status: %d (%s)\n", osr,
               (osr ? "line enabled as output" : "line pure input"));

    mmap_ptr = check_mmap(mem_fd, pio_ifsr[pioc_num], &mstat);
    if (NULL == mmap_ptr)
        return 1;
    ap = mmap_ptr + (pio_ifsr[pioc_num] & MAP_MASK);
    ifsr = !!( *((unsigned int *)ap) & bit_mask);
    if (0 == brief)
        printf("  input filter status: %d (%s)\n", ifsr,
               (ifsr ? "glitch filter ENabled" : "glitch filter DISabled"));

    mmap_ptr = check_mmap(mem_fd, pio_odsr[pioc_num], &mstat);
    if (NULL == mmap_ptr)
        return 1;
    ap = mmap_ptr + (pio_odsr[pioc_num] & MAP_MASK);
    odsr = !!( *((unsigned int *)ap) & bit_mask);
    if (0 == brief)
        printf("  output data status: %d\n", odsr);

    mmap_ptr = check_mmap(mem_fd, pio_pdsr[pioc_num], &mstat);
    if (NULL == mmap_ptr)
        return 1;
    ap = mmap_ptr + (pio_pdsr[pioc_num] & MAP_MASK);
    pdsr = !!( *((unsigned int *)ap) & bit_mask);
    if (0 == brief)
        printf("  pin data status: %d\n", pdsr);

    mmap_ptr = check_mmap(mem_fd, pio_imr[pioc_num], &mstat);
    if (NULL == mmap_ptr)
        return 1;
    ap = mmap_ptr + (pio_imr[pioc_num] & MAP_MASK);
    imr = !!( *((unsigned int *)ap) & bit_mask);
    if (0 == brief)
        printf("  interrupt mask: %d (%s)\n", imr,
               (imr ? "ENabled" : "DISabled"));

    if (interrupt) {
        mmap_ptr = check_mmap(mem_fd, pio_isr[pioc_num], &mstat);
        if (NULL == mmap_ptr)
            return 1;
        ap = mmap_ptr + (pio_isr[pioc_num] & MAP_MASK);
        isr = !!( *((unsigned int *)ap) & bit_mask);
        if (0 == brief)
            printf("  interrupt status: %d (%s)\n", isr,
                   (isr ? "input CHANGE" : "NO input change"));
    } else
        isr = -1;

    mmap_ptr = check_mmap(mem_fd, pio_mdsr[pioc_num], &mstat);
    if (NULL == mmap_ptr)
        return 1;
    ap = mmap_ptr + (pio_mdsr[pioc_num] & MAP_MASK);
    mdsr = !!( *((unsigned int *)ap) & bit_mask);
    if (0 == brief)
        printf("  multi drive status: %d (%s)\n", mdsr,
               (mdsr ? "enabled, pin driven when low" :
               "pin driven high and low"));

    mmap_ptr = check_mmap(mem_fd, pio_pusr[pioc_num], &mstat);
    if (NULL == mmap_ptr)
        return 1;
    ap = mmap_ptr + (pio_pusr[pioc_num] & MAP_MASK);
    pusr = !!( *((unsigned int *)ap) & bit_mask);
    if (0 == brief)
        printf("  pull-up status: %d (%s)\n", pusr,
               (pusr ? "DISabled" : "ENabled"));

    if (psr) {
        absr = -1;
        mmap_ptr = check_mmap(mem_fd, pio_owsr[pioc_num], &mstat);
        if (NULL == mmap_ptr)
            return 1;
        ap = mmap_ptr + (pio_owsr[pioc_num] & MAP_MASK);
        owsr = !!( *((unsigned int *)ap) & bit_mask);
        if (0 == brief)
            printf("  output write status: %d (%s)\n", owsr,
                   (owsr ? "writing PIO_ODSR affects IO line" :
                           "writing PIO_ODSR ignored"));
    } else {
        /* peripheral mode */
        owsr = -1;
        mmap_ptr = check_mmap(mem_fd, pio_absr[pioc_num], &mstat);
        if (NULL == mmap_ptr)
            return 1;
        ap = mmap_ptr + (pio_absr[pioc_num] & MAP_MASK);
        absr = !!( *((unsigned int *)ap) & bit_mask);
        if (0 == brief)
            printf("  peripheral A B status: %d (%s)\n", absr,
                   (absr ? "B" : "A"));
    }
    if (brief) {
        if (1 == brief)
            printf("psr=%d osr=%d ifsr=%d odsr=%d pdsr=%d imr=%d isr=%d "
                   "mdsr=%d pusr=%d absr=%d owsr=%d\n", psr, osr, ifsr,
                   odsr, pdsr, imr, isr, mdsr, pusr, absr, owsr);
        else
            printf("%-2d: psr=%d osr=%d ifsr=%d odsr=%d pdsr=%d "
                   "mdsr=%d pusr=%d absr=%d\n", bit_num, psr, osr, ifsr,
                   odsr, pdsr, mdsr, pusr, absr);
    }

    if (mstat.mmap_ok) {
        if (-1 == munmap(mmap_ptr, MAP_SIZE)) {
            fprintf(stderr, "mmap_ptr=%p:\n", mmap_ptr);
            perror("    munmap");
            return 1;
        } else if (verbose > 2)
            fprintf(stderr, "trailing munmap() ok, mmap_ptr=%p\n", mmap_ptr);
    }
    if (mem_fd >= 0)
        close(mem_fd);
    return 0;
}
